#include <algorithm>        // for swap
#include <memory>           // for unique_ptr
#include <cstdlib> // for posix_memalign / aligned_alloc / malloc; free
#include <cstring>          // for memcmp, memcpy


using char8 = signed char;
using uchar8 = unsigned char;
using uint32 = unsigned int;
using int64 = long long;
using uint64 = unsigned long long;
using int32 = signed int;
using ushort16 = unsigned short;
using short16 = signed short;

// This allows to specify the nuber of bytes that each Buffer needs to
// allocate additionally to be able to remove one runtime bounds check
// in BitStream::fill. There are two sane choices:
// 0 : allocate exactly as much data as required, or
// set it to the value of  BitStreamCacheBase::MaxProcessBytes
#define BUFFER_PADDING 0UL

// if the padding is >= 4, bounds checking in BitStream::fill are not compiled,
// which supposedly saves about 1% on modern CPUs
// WARNING: if the padding is >= 4, do *NOT* create Buffer from
// passed unowning pointer and size. Or, subtract BUFFER_PADDING from size.
// else bound checks will malfunction => bad things can happen !!!

/*************************************************************************
 * This is the buffer abstaction.
 *
 * It allows access to some piece of memory, typically a whole or part
 * of a raw file. The underlying memory may be owned by the buffer or not.
 * It supports move operations to properly deal with owneship transfer.
 * It intentionally supports only read/const access to the underlying memory.
 *
 *************************************************************************/
class Buffer
{
public:
  using size_type = uint32;

  // allocates the databuffer, and returns owning non-const pointer.
  static std::unique_ptr<uchar8, decltype(&free)> Create(size_type size);

  // constructs an empty buffer
  Buffer() = default;
  // creates buffer from owning unique_ptr
  Buffer(std::unique_ptr<uchar8, decltype(&free)> data_,
         size_type size_);

  // Allocates the memory
  explicit Buffer(size_type size);

  // Data already allocated
  explicit Buffer(const uchar8* data_, size_type size_)
      : data(data_), size(size_) {
    static_assert(BUFFER_PADDING == 0, "please do make sure that you do NOT "
                                       "call this function from YOUR code, and "
                                       "then comment-out this assert.");
  }
  // creates a (non-owning) copy / view of rhs
  Buffer(const Buffer& rhs)
    : data(rhs.data), size(rhs.size) {}
  // Move data and ownership from rhs to this
  Buffer(Buffer&& rhs) noexcept
    : data(rhs.data), size(rhs.size), isOwner(rhs.isOwner) { rhs.isOwner = false; }
  // Frees memory if owned
  ~Buffer();
  Buffer& operator=(const Buffer& rhs);

  Buffer getSubView(size_type offset, size_type size_) const {
    return Buffer(getData(offset, size_), size_);
  }
  Buffer getSubView(size_type offset) const {

    size_type newSize = size - offset;
    return Buffer(getData(offset, newSize), newSize);
  }

  // get pointer to memory at 'offset', make sure at least 'count' bytes are accessable
  const uchar8* getData(size_type offset, size_type ) const {

    return &data[offset];
  }

  // convenience getter for single bytes
  uchar8 operator[](size_type offset) const {
    return *getData(offset, 1);
  }

  // std begin/end iterators to allow for range loop
  const uchar8* begin() const {
    return data;
  }
  const uchar8* end() const {
    return data + size;
  }


  inline size_type getSize() const {
    return size;
  }

  inline bool isValid(size_type offset, size_type count = 1) const {
    return (uint64)offset + count <= (uint64)size + BUFFER_PADDING;
  }

//  Buffer* clone();
//  /* For testing purposes */
//  void corrupt(int errors);
//  Buffer* cloneRandomSize();

protected:
  const uchar8* data = nullptr;
  size_type size = 0;
  bool isOwner = false;
};

/*
 * DataBuffer is a simple extention to Buffer. It knows about the byte order
 * of its contents and can therefore provide save access to larger than
 * byte sized data, like int, float, etc.
 */
class DataBuffer : public Buffer
{
  bool inNativeByteOrder = true;
public:
  DataBuffer() = default;
  explicit DataBuffer(const Buffer& data_, bool inNativeByteOrder_ = true)
      : Buffer(data_), inNativeByteOrder(inNativeByteOrder_) {}

  inline bool isInNativeByteOrder() const {
    return inNativeByteOrder;
  }

  inline bool setInNativeByteOrder(bool value) {
    std::swap(inNativeByteOrder, value);
    return value;
  }
};

class ByteStream : public DataBuffer
{
protected:
  size_type pos = 0; // position of stream in bytes (this is next byte to deliver)

public:
  ByteStream() = default;
  explicit ByteStream(const DataBuffer& buffer) : DataBuffer(buffer) {}
  ByteStream(const Buffer &buffer, size_type offset, size_type size_,
             bool inNativeByteOrder_ = true)
      : DataBuffer(buffer.getSubView(0, offset + size_), inNativeByteOrder_),
        pos(offset) {}
  ByteStream(const Buffer &buffer, size_type offset,
             bool inNativeByteOrder_ = true)
      : DataBuffer(buffer, inNativeByteOrder_), pos(offset) {}

  // deprecated:
  ByteStream(const Buffer* f, size_type offset, size_type size_,
             bool inNativeByteOrder_ = true)
      : ByteStream(*f, offset, size_, inNativeByteOrder_) {}
  ByteStream(const Buffer* f, size_type offset, bool inNativeByteOrder_ = true)
      : ByteStream(*f, offset, inNativeByteOrder_) {}

  // return ByteStream that starts at given offset
  // i.e. this->data + offset == getSubStream(offset).data
  ByteStream getSubStream(size_type offset, size_type size_) {
    return ByteStream(getSubView(offset, size_), 0, isInNativeByteOrder());
  }

  inline void check(size_type ) const {
  }

  inline size_type getPosition() const { return pos; }
  inline void setPosition(size_type newPos) {
    pos = newPos;
    check(0);
  }
  inline size_type getRemainSize() const { return size-pos; }
  inline const uchar8* peekData(size_type count) {
    return Buffer::getData(pos, count);
  }
  inline const uchar8* getData(size_type count) {
    const uchar8* ret = Buffer::getData(pos, count);
    pos += count;
    return ret;
  }
  inline Buffer getBuffer(size_type size_) {
    Buffer ret = getSubView(pos, size_);
    pos += size_;
    return ret;
  }

  inline uchar8 peekByte(size_type i = 0) const {
    check(i+1);
    return data[pos+i];
  }

  inline void skipBytes(size_type nbytes) {
    pos += nbytes;
    check(0);
  }

  inline bool hasPatternAt(const char *pattern, size_type size_,
                           size_type relPos) const {
    if (!isValid(pos + relPos, size_))
      return false;
    return memcmp(&data[pos + relPos], pattern, size_) == 0;
  }

  inline bool hasPrefix(const char *prefix, size_type size_) const {
    return hasPatternAt(prefix, size_, 0);
  }

  inline bool skipPrefix(const char *prefix, size_type size_) {
    bool has_prefix = hasPrefix(prefix, size_);
    if (has_prefix)
      pos += size_;
    return has_prefix;
  }

  inline uchar8 getByte() {
    check(1);
    return data[pos++];
  }

  const char* peekString() const {
    size_type p = pos;
    do {
      check(1);
    } while (data[p++] != 0);
    return (const char*)&data[pos];
  }

  // Increments the stream to after the next zero byte and returns the bytes in between (not a copy).
  // If the first byte is zero, stream is incremented one.
  const char* getString() {
    size_type start = pos;
    do {
      check(1);
    } while (data[pos++] != 0);
    return (const char*)&data[start];
  }

  // recalculate the internal data/position information such that current position
  // i.e. getData() before == getData() after but getPosition() after == newPosition
  // this is only used for DNGPRIVATEDATA handling to restore the original offset
  // in case the private data / maker note has been moved within in the file
  // TODO: could add a lower bound check later if required.
  void rebase(size_type newPosition, size_type newSize) {
    const uchar8* dataAtNewPosition = getData(newSize);
    data = dataAtNewPosition - newPosition;
    size = newPosition + newSize;
  }

  // special factory function to set up internal buffer with copy of passed data.
  // only necessary to create 'fake' TiffEntries (see e.g. RAF)
  static ByteStream createCopy(void* data, size_type size) {
    ByteStream bs;
    auto* new_data = (uchar8*)malloc(size);
    memcpy(new_data, data, size);
    bs.data = new_data;
    bs.size = size;
    bs.isOwner = true;
    return bs; // hint: copy elision or move will happen
  }
};

std::unique_ptr<uchar8, decltype(&free)> Buffer::Create(size_type size) {
  std::unique_ptr<uchar8, decltype(&free)> data(
      (uchar8*)malloc(size + BUFFER_PADDING),
      free);

  return data;
}

Buffer::Buffer(std::unique_ptr<uchar8, decltype(&free)> data_,
               size_type size_)
    : size(size_) {

  data = data_.release();

  isOwner = true;
}

Buffer::Buffer(size_type size_) : Buffer(Create(size_), size_) {}

Buffer::~Buffer() {
  if (isOwner) {
    free(const_cast<uchar8*>(data));
  }
}

Buffer& Buffer::operator=(const Buffer &rhs)
{
  this->~Buffer();
  data = rhs.data;
  size = rhs.size;
  isOwner = false;
  return *this;
}

#define JPEG_MEMSRC(B, C) jpeg_mem_src_int(B, C)
static int jpeg_mem_src_int(const unsigned char* , int nbytes) {
  return nbytes;
}

int main() {
  ByteStream input(DataBuffer(Buffer(10)));

  return JPEG_MEMSRC(input.getData(input.getRemainSize()), input.getRemainSize());
}
